/****************************************************************
 * C API to the LegoUSB Tower driver.
 * User-space applications may call these routines to configure
 * a USB Tower.
 *
 * Copyright (c) 2003 Stewart Heitmann <sheitmann@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 ****************************************************************/

#include "legousbapi.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <errno.h>
#include "driver.h"

int legousbapi_debug = 0;

static void print_error(int err)
	{
	fprintf(stderr,"error(0x%02x): %s\n",(-err),legousbapi_strerror(err));
	}

char* legousbapi_strerror(int err)
	{
    switch(-err)
		{
		case LTW_REQERR_SUCCESS:
			return "Success";
	    case LTW_REQERR_BADPARM:
			return "Bad parameter or value";
		case LTW_REQERR_BUSY:
		    return "Tower is busy";
		case LTW_REQERR_NOPOWER:
		    return "Insufficient power";
		case LTW_REQERR_WRONGMODE:
		    return "Tower is in wrong mode for request";
		case LTW_INTERNAL_ERROR:
		    return "Tower had internal error";
		case LTW_REQERR_BADREQUEST:
		    return "Bad request";
		}
	return "Unknown error";
	}

int legousbapi_set_parm(int fd, int param, int value)
	{
	const int req = value<<8 | param;
    const int err = ioctl(fd, LEGO_TOWER_SET_PARAM, req);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_SET_PARAM 0x%02x ",req);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}

int legousbapi_get_parm(int fd, int param)
	{
    const int err = ioctl(fd, LEGO_TOWER_GET_PARAM, param);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_PARAM 0x%02x returns ",param);
		if (err<0)
			print_error(err);
		else
			{
			switch(param)
				{
				case LTW_PARM_MODE:
					switch(err)
						{
						case LTW_MODE_VLL:
							fprintf(stderr,"MODE_VLL\n");
							break;
						case LTW_MODE_IR:
							fprintf(stderr,"MODE_IR\n");
							break;
						case LTW_MODE_IRC:
							fprintf(stderr,"MODE_IRC\n");
							break;
						case LTW_MODE_RADIO:
							fprintf(stderr,"MODE_RADIO\n");
							break;
						default:
							fprintf(stderr,"0x%02x\n",err);
							break;
						}
					break;

				case LTW_PARM_RANGE:
					switch(err)
						{
						case LTW_RANGE_SHORT:
							fprintf(stderr,"RANGE_SHORT\n");
							break;
						case LTW_RANGE_MEDIUM:
							fprintf(stderr,"RANGE_MEDIUM\n");
							break;
						case LTW_RANGE_LONG:
							fprintf(stderr,"RANGE_LONG\n");
							break;
						default:
							fprintf(stderr,"0x%02x\n",err);
							break;
						}
					break;

				case LTW_PARM_ERRDETECT:
					switch(err)
						{
						case LTW_ERRDETECT_ON:
							fprintf(stderr,"ERRDETECT_ON\n");
							break;
						case LTW_ERRDETECT_OFF:
							fprintf(stderr,"ERRDETECT_OFF\n");
							break;
						default:
							fprintf(stderr,"0x%02x\n",err);
							break;
						}
					break;

				case LTW_PARM_ERRSTATUS:
					switch(err)
						{
						case LTW_NO_ERROR:
							fprintf(stderr,"NO_ERROR\n");
							break;
						case LTW_NO_SUPPORT:
							fprintf(stderr,"NO_SUPPORT\n");
							break;
						case LTW_ERR_VEND_REQ:
							fprintf(stderr,"ERR_VEND_REQ\n");
							break;
						case LTW_ERR_USB:
							fprintf(stderr,"ERR_USB\n");
							break;
						case LTW_ERR_USB_FUNC:
							fprintf(stderr,"ERR_USB_FUNC\n");
							break;
						case LTW_ERR_USB_DATA:
							fprintf(stderr,"ERR_USB_DATA\n");
							break;
						case LTW_ERR_USB_ENDPOINT:
							fprintf(stderr,"ERR_USB_ENDPOINT\n");
							break;
						case LTW_ERR_USB_PACKETSIZE:
							fprintf(stderr,"ERR_USB_PACKETSIZE\n");
							break;
						case LTW_ERR_USB_TX_EP0:
							fprintf(stderr,"ERR_USB_TX_EP0\n");
							break;
						case LTW_ERR_USB_EP0_PACKETSIZE:
							fprintf(stderr,"ERR_USB_EP0_PACKETSIZE\n");
							break;
						case LTW_ERR_USB_EP1_PACKETSIZE:
							fprintf(stderr,"ERR_USB_EP1_PACKETSIZE\n");
							break;
						case LTW_ERR_USB_VEND_BUFF_FULL:
							fprintf(stderr,"ERR_USB_VEND_BUFF_FULL\n");
							break;
						case LTW_ERR_USB_TX_EP1:
							fprintf(stderr,"ERR_USB_TX_EP1\n");
							break;
						case LTW_ERR_UART:
							fprintf(stderr,"ERR_UART\n");
							break;
						case LTW_ERR_UART_MODE:
							fprintf(stderr,"ERR_UART_MODE\n");
							break;
						case LTW_ERR_UART_TX_STATE:
							fprintf(stderr,"ERR_UART_TX_STATE\n");
							break;
						case LTW_ERR_UART_LOW_POWER:
							fprintf(stderr,"ERR_UART_LOW_POWER\n");
							break;
						case LTW_ERR_TOWER:
							fprintf(stderr,"ERR_TOWER\n");
							break;
						case LTW_ERR_UNDEFINED:
							fprintf(stderr,"ERR_UNDEFINED\n");
							break;
						default:
							fprintf(stderr,"0x%02x\n",err);
							break;
						}
					break;

				case LTW_PARM_ENDIAN:
					switch(err)
						{
						case LTW_ENDIAN_LITTLE:
							fprintf(stderr,"ENDIAN_LITTLE\n");
							break;
						case LTW_ENDIAN_BIG:
							fprintf(stderr,"ENDIAN_BIG\n");
							break;
						default:
							fprintf(stderr,"0x%02x\n",err);
							break;
						}
					break;

				case LTW_PARM_ID_LED_MODE:
					switch(err)
						{
						case LTW_ID_LED_HW_CTRL:
							fprintf(stderr,"ID_LED_HW_CTRL\n");
							break;
						case LTW_ID_LED_SW_CTRL:
							fprintf(stderr,"ID_LED_SW_CTRL\n");
							break;
						default:
							fprintf(stderr,"0x%02x\n",err);
							break;
						}
					break;

				case LTW_PARM_ERROR_SIGNAL:
					switch(err)
						{
						case LTW_ID_LED_ON:
							fprintf(stderr,"ID_LED_ON\n");
							break;
						case LTW_ID_LED_OFF:
							fprintf(stderr,"ID_LED_OFF\n");
							break;
						default:
							fprintf(stderr,"0x%02x\n",err);
							break;
						}
					break;

				default:
					fprintf(stderr,"0x%02x\n",err);
					break;
				};
			};
		}
	return err;
	}

//I suspect a bug in the driver.
//It does not return buffno in LTW_REQ_FLUSH_REPLY.bBufferno
//as it should.
int legousbapi_flush(int fd, int buffno)
	{
    const int err = ioctl(fd, LEGO_TOWER_FLUSH, buffno);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_FLUSH 0x%02x ",buffno);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}

int legousbapi_reset(int fd)
	{
    const int err = ioctl(fd, LEGO_TOWER_RESET, 0);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_RESET ");
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}

int legousbapi_get_stat(int fd, struct legousbapi_get_stat_reply *stat)
    {
	LTW_REQ_GET_STAT_REPLY reply;
    const int err = ioctl(fd, LEGO_TOWER_GET_STAT, &reply);
	if (err<0)
		memset(stat,0,sizeof(struct legousbapi_get_stat_reply));
	else
		{
		stat->NoOfRxBytes       = (int)reply.wNoOfRxBytes;
		stat->OverrunErrorCount = (int)reply.wOverrunErrorCount;
		stat->NoiseCount        = (int)reply.wNoiseCount;
		stat->FramingErrorCount = (int)reply.wFramingErrorCount;
		}
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_STAT returns ");
		if (err<0)
			print_error(err);
		else
			{
	    	fprintf(stderr,"ok(0x%02x)\n",err);
	    	fprintf(stderr,"NoOfRxBytes=%d OverrunErrorCount=%d NoiseCount=%d FramingErrorCount=%d\n",
 		            stat->NoOfRxBytes, stat->OverrunErrorCount, stat->NoiseCount, stat->FramingErrorCount);
			}
		}
	return err;
	}

int legousbapi_get_power(int fd)
	{
	const int err = ioctl(fd, LEGO_TOWER_GET_POWER, 0);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_POWER returns ");
		switch(err)
			{
			case LTW_POWER_LOW:
				fprintf(stderr,"POWER_LOW\n");
				break;
			case LTW_POWER_HIGH:
				fprintf(stderr,"POWER_HIGH\n");
				break;
			default:
				print_error(err);
				break;
			}
		}
	return err;
	}

int legousbapi_get_led(int fd, int led_id)
	{
    const int err = ioctl(fd, LEGO_TOWER_GET_LED, led_id);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_LED 0x%02x returns ",led_id);
		switch(err)
			{
			case LTW_LED_COLOR_BLACK:
				fprintf(stderr,"LED_COLOR_BLACK\n");
				break;
			case LTW_LED_COLOR_GREEN:
				fprintf(stderr,"LED_COLOR_GREEN\n");
				break;
			case LTW_LED_COLOR_YELLOW:
				fprintf(stderr,"LED_COLOR_YELLOW\n");
				break;
			case LTW_LED_COLOR_ORANGE:
				fprintf(stderr,"LED_COLOR_ORANGE\n");
				break;
			case LTW_LED_COLOR_RED:
				fprintf(stderr,"LED_COLOR_RED\n");
				break;
			default:
	  		    print_error(err);
				break;
			}
		}
	return err;
	}

int legousbapi_set_led(int fd, int led_id, int led_color)
	{
	const int req = led_color<<8 | led_id;
	const int err = ioctl(fd, LEGO_TOWER_SET_LED, req);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_SET_LED 0x%02x ",req);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}

int legousbapi_reset_stat(int fd)
    {
    const int err = ioctl(fd, LEGO_TOWER_RESET_STAT, 0);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_RESET_STAT ");
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}

int legousbapi_get_parm_irc(int fd, int param)
	{
    const int err = ioctl(fd, LEGO_TOWER_GET_PARAM_IRC, param);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_PARAM_IRC 0x%02x returns ",param);
		if (err<0)
			print_error(err);
		else
			{
			switch(param)
				{
				case LTW_PARM_IRC_PACKETSIZE:
					fprintf(stderr,"%d bytes\n",err);
					break;
				case LTW_PARM_IRC_TIMEFRAME:
					fprintf(stderr,"%d msecs\n",err);
					break;
				default:
					fprintf(stderr,"%d\n",err);
					break;
				};
			};
		}
	return err;
	}

int legousbapi_set_parm_irc(int fd, int param, int value)
	{
	const int req = value<<8 | param;
    const int err = ioctl(fd, LEGO_TOWER_SET_PARAM_IRC, req);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_SET_PARAM_IRC 0x%02x ",req);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}


int legousbapi_get_tx_speed(int fd)
	{
    const int err = ioctl(fd, LEGO_TOWER_GET_TX_SPEED, 0);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_TX_SPEED returns ");
		switch(err)
			{
			case SPEED_COMM_BAUD_1200:
				fprintf(stderr,"SPEED_COMM_BAUD_1200\n");
				break;
			case SPEED_COMM_BAUD_2400:
				fprintf(stderr,"SPEED_COMM_BAUD_2400\n");
				break;
			case SPEED_COMM_BAUD_4800:
				fprintf(stderr,"SPEED_COMM_BAUD_4800\n");
				break;
			case SPEED_COMM_BAUD_9600:
				fprintf(stderr,"SPEED_COMM_BAUD_9600\n");
				break;
			case SPEED_COMM_BAUD_19200:
				fprintf(stderr,"SPEED_COMM_BAUD_19200\n");
				break;
			default:
				print_error(err);
				break;
			}
		}
	return err;
	}

int legousbapi_set_tx_speed(int fd, int tx_speed)
	{
    const int err = ioctl(fd, LEGO_TOWER_SET_TX_SPEED, tx_speed);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_SET_TX_SPEED 0x%02x ",tx_speed);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}

int legousbapi_get_rx_speed(int fd)
	{
    const int err = ioctl(fd, LEGO_TOWER_GET_RX_SPEED, 0);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_RX_SPEED returns ");
		switch(err)
			{
			case SPEED_COMM_BAUD_1200:
				fprintf(stderr,"SPEED_COMM_BAUD_1200\n");
				break;
			case SPEED_COMM_BAUD_2400:
				fprintf(stderr,"SPEED_COMM_BAUD_2400\n");
				break;
			case SPEED_COMM_BAUD_4800:
				fprintf(stderr,"SPEED_COMM_BAUD_4800\n");
				break;
			case SPEED_COMM_BAUD_9600:
				fprintf(stderr,"SPEED_COMM_BAUD_9600\n");
				break;
			case SPEED_COMM_BAUD_19200:
				fprintf(stderr,"SPEED_COMM_BAUD_19200\n");
				break;
			default:
				print_error(err);
				break;
			}
		}
	return err;
	}

int legousbapi_set_rx_speed(int fd, int rx_speed)
	{
    const int err = ioctl(fd, LEGO_TOWER_SET_RX_SPEED, rx_speed);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_SET_RX_SPEED 0x%02x ",rx_speed);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}

int legousbapi_get_tx_state(int fd)
    {
    const int err = ioctl(fd, LEGO_TOWER_GET_TX_STATE,0);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_TX_STATE returns ");
		switch(err)
			{
			case CAPS_TX_STATE_READY:
				fprintf(stderr,"CAPS_TX_STATE_READY\n");
				break;
			case CAPS_TX_STATE_BUSY:
				fprintf(stderr,"CAPS_TX_STATE_BUSY\n");
				break;
			default:
				print_error(err);
				break;
			};
		}
	return err;
	}

int legousbapi_get_tx_carrier_frequency(int fd)
	{
	const int err  = ioctl(fd, LEGO_TOWER_GET_TX_CARRIER_FREQUENCY, 0);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_TX_CARRIER_FREQUENCY returns ");
		if (err<0)
			print_error(err);
		else
		    fprintf(stderr,"%dkHz\n",err);
		}
	return err;
	}

int legousbapi_set_tx_carrier_frequency(int fd, int carrier_frequency)
	{
    const int err = ioctl(fd, LEGO_TOWER_SET_TX_CARRIER_FREQUENCY, carrier_frequency);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_SET_TX_CARRIER_FREQUENCY 0x%02x ",carrier_frequency);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}

int legousbapi_get_tx_carrier_duty_cycle(int fd, int tx_range)
	{
    const int err = ioctl(fd, LEGO_TOWER_GET_TX_CARRIER_DUTY_CYCLE, tx_range);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_TX_CARRIER_DUTY_CYCLE 0x%02x returns ",tx_range);
		if (err<0)
			print_error(err);
		else
			fprintf(stderr," %d x 10e-7 secs\n",err);
		}
	return err;
	}

int legousbapi_set_tx_carrier_duty_cycle(int fd, int tx_range, int duty_cycle)
	{
	//const int req = duty_cycle<<16 | (tx_range & 0xffff);
	const int req = tx_range<<16 | (duty_cycle & 0xffff);
	const int err = ioctl(fd, LEGO_TOWER_SET_TX_CARRIER_DUTY_CYCLE, req);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_SET_TX_CARRIER_DUTY_CYCLE 0x%02x ",req);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}

int legousbapi_get_caps(int fd, int mode, struct legousbapi_get_caps_reply *caps)
    {
	int err;
	union
	   {
	   int request;
	   LTW_REQ_GET_CAPS_REPLY reply;
	   } data;

	data.request=mode;
    err = ioctl(fd, LEGO_TOWER_GET_CAPS, &data);
	if (err<0)
		memset(caps,0,sizeof(*caps));
	else
		{
		caps->CommDirections=(int)data.reply.bCommDirections;
		caps->CommRange=(int)data.reply.bCommRange;
		caps->CommTransmitRate=(int)data.reply.wCommTransmitRate;
		caps->CommReceiveRate=(int)data.reply.wCommReceiveRate;
		caps->CommTransmitMinFrequency=(int)data.reply.bCommTransmitMinFrequency;
		caps->CommTransmitMaxFrequency=(int)data.reply.bCommTransmitMaxFrequency;
		caps->CommTransmitMinDutyCycle=(int)data.reply.wCommTransmitMinDutyCycle;
		caps->CommTransmitMaxDutyCycle=(int)data.reply.wCommTransmitMaxDutyCycle;
		caps->NoOfBytesInUartTxBuffer=(int)data.reply.bNoOfBytesInUartTxBuffer;
		caps->NoOfBytesInUartRxBuffer=(int)data.reply.bNoOfBytesInUartRxBuffer;
		}
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_CAPS 0x%02x ", mode);
		if (err<0)
			print_error(err);
		else
			{
	    	fprintf(stderr,"ok(0x%02x)\n",err);
			fprintf(stderr,"CommDirections=0x%02x\n",   caps->CommDirections);
			fprintf(stderr,"CommRange=0x%02x\n",        caps->CommRange);
			fprintf(stderr,"CommTransmitRate=0x%02x\n", caps->CommTransmitRate);
			fprintf(stderr,"CommReceiveRate=0x%02x\n",  caps->CommReceiveRate);
			fprintf(stderr,"CommTransmitMinFrequency=%d\n", caps->CommTransmitMinFrequency);
			fprintf(stderr,"CommTransmitMaxFrequency=%d\n", caps->CommTransmitMaxFrequency);
			fprintf(stderr,"CommTransmitMinDutyCycle=%d\n", caps->CommTransmitMinDutyCycle);
			fprintf(stderr,"CommTransmitMaxDutyCycle=%d\n", caps->CommTransmitMaxDutyCycle);
			fprintf(stderr,"NoOfBytesInUartTxBuffer=%d\n",  caps->NoOfBytesInUartTxBuffer);
			fprintf(stderr,"NoOfBytesInUartRxBuffer=%d\n",  caps->NoOfBytesInUartRxBuffer);
			}
		}
	return err;
	}

int legousbapi_get_version(int fd, struct legousbapi_get_version_reply *version)
	{
	LTW_REQ_GET_VERSION_REPLY reply;
    const int err = ioctl(fd, LEGO_TOWER_GET_VERSION, &reply);
	if (err<0)
		memset(version,0,sizeof(struct legousbapi_get_version_reply));
	else
		{
		version->MajorVersion = (int)reply.bMajorVersion;
		version->MinorVersion = (int)reply.bMinorVersion;
		version->BuildNo      = (int)reply.wBuildNo;
		}
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_VERSION ");
		if (err<0)
			print_error(err);
		else
			{
	    	fprintf(stderr,"ok(0x%02x)\n",err);
			fprintf(stderr,"MajorVersion=%d ", version->MajorVersion);
			fprintf(stderr,"MinorVersion=%d ", version->MinorVersion);
			fprintf(stderr,"BuildNo=%d\n",     version->BuildNo);
			}
		}
	return err;
	}


int legousbapi_get_copyright(int fd, char* copyright, int n)
    {
	int err;
	union
	   {
	   int request;
	   struct
	      {
	      LTW_REQ_GET_COPYRIGHT_REPLY reply;
	      Word unicode[128];
	      } reply;
	    } data;

    data.request = (int)sizeof(data);
    err = ioctl(fd, LEGO_TOWER_GET_COPYRIGHT, &data);
 	if (err<0)
		copyright[0]='\0';
	else
	    {
		int i;
		int nwords = (err-4)/2;   /* (nbytes-replyheader)/sizeof(word) */
		int nchar=0;
		for (i=0; i<nwords; i++)
			{
			/* convert each unicode word to ascii char */
			wchar_t wc = (wchar_t)data.reply.unicode[i];
			if (nchar+MB_CUR_MAX>=n)
				break;
			nchar += wctomb(copyright+nchar,wc);
			}
		copyright[nchar++]='\0';
		}
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_COPYRIGHT ");
		if (err<0)
			print_error(err);
		else
			fprintf(stderr,"\n%s",copyright);
		}
	return err;
	}


int legousbapi_get_credits(int fd, char* credits, int n)
    {
	int err;
	union
	   {
	   int request;
	   struct
	      {
	      LTW_REQ_GET_CREDITS_REPLY reply;
	      Word unicode[0x500];
	      } reply;
	    } data;

	//memset(credits,0,n);

    data.request = (int)sizeof(data);
    err = ioctl(fd, LEGO_TOWER_GET_CREDITS, &data);
 	if (err<0)
		credits[0]='\0';
	else
	    {
		int i;
		int nwords = (err-4)/2;   /* (nbytes-replyheader)/sizeof(word) */
		int nchar=0;
		for (i=0; i<nwords; i++)
			{
			/* convert each unicode word to ascii char */
			wchar_t wc = (wchar_t)data.reply.unicode[i];
			if (nchar+MB_CUR_MAX>=n)
				break;
			nchar += wctomb(credits+nchar,wc);
			//fprintf(stderr,"wc[%d]=%ld, credits[%d]=0x%02x\n",i,wc,nchar,credits+nchar);
			}
		credits[nchar++]='\0';
		}
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_GET_CREDITS ");
		if (err<0)
			print_error(err);
		else
			fprintf(stderr,"\n%s",credits);
		}
	return err;
	}


int legousbapi_set_read_timeout(int fd, int timeout)
	{
    const int err = ioctl(fd, LEGO_TOWER_SET_READ_TIMEOUT, timeout);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_SET_READ_TIMEOUT 0x%02x ", timeout);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}


int legousbapi_set_write_timeout(int fd, int timeout)
	{
    const int err = ioctl(fd, LEGO_TOWER_SET_WRITE_TIMEOUT, timeout);
	if (legousbapi_debug)
		{
        fprintf(stderr,"LEGO_TOWER_SET_WRITE_TIMEOUT 0x%02x ", timeout);
		if (err<0)
			print_error(err);
		else
	    	fprintf(stderr,"ok(0x%02x)\n",err);
		}
	return err;
	}




