/*! \file   command_server.c
    \brief  Command server for ULeague
    \author Jacky Baltes <jacky@cs.umanitoba.ca>
*/

/*
 *  The contents of this file are subject to the Mozilla Public License
 *  Version 1.0 (the "License"); you may not use this file except in
 *  compliance with the License. You may obtain a copy of the License at
 *  http://www.mozilla.org/MPL/
 *
 *  Software distributed under the License is distributed on an "AS IS"
 *  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 *  License for the specific language governing rights and limitations
 *  under the License.
 *
 *  The Original Code is legOS code, released October 2, 1999.
 *
 *  The Initial Developer of the Original Code is Markus L. Noga.
 *  Portions created by Markus L. Noga are Copyright (C) 1999
 *  Markus L. Noga. All Rights Reserved.
 *
 *  Contributor(s): everyone discussing LNP at LUGNET
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <ctype.h>
#include <pthread.h>
#include <scanner.h>
#include <command_codes.h>
#include <math.h>
#include <legousbapi.h>

#if defined(_WIN32)
  #include <windows.h>
#endif

#if (defined(__unix__) || defined(unix)) && !defined(USG)
#include <sys/param.h>
#endif

#include <sys/lnp.h>
#include <sys/lnp-logical.h>

#include "rcxtty.h"

#define DEFAULT_PORT   8765

#if (defined(__sun__) && defined(__svr4__)) || defined(BSD)	// Solaris||BSD
#undef HAVE_GETOPT_LONG
#else
#define HAVE_GETOPT_LONG 1
#endif

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>

static const struct option long_options[]={
  {"tty",    required_argument,0,'t'},
  {"irmode", required_argument,0,'i'},
  {"port", required_argument,0,'p'},
  {"robot", required_argument,0,'r'},
  {"verbose",no_argument      ,0,'v'},
  {0        ,0                ,0,0  }
};

#else // HAVE_GETOPT_LONG

#define getopt_long(ac, av, opt, lopt, lidx) (getopt((ac), (av), (opt)))

#endif // HAVE_GETOPT_LONG

extern char * yytext;
int yylex();

struct RobotMapping 
{
  char name[64];
  int  id;
};

struct RobotMapping robotMaps[32];

char * messageBuffer;
int messageLength;

static unsigned char volatile currentCommand[8];

/* Prototypes */
int extractRobotMapping( char const * arg, char * name, int * id );
int insertRobotMapping( struct RobotMapping * maps, int numMaps, char * name, int id );
int lnp_logical_write(const void *data, size_t length); 
void LNPinit(const char *tty);
int initUDPBroadcastSocket( int port );
void listener( int * s );
int findRobotMapping( char const * name );
int parseMessage( char * buffer, int len );
int sendMessage( unsigned char volatile * buffer, int len );
int addPort( int ports[], int numPorts, char * newPort );
unsigned int calcCheckSum( unsigned char * message, int length );
unsigned char encodeCommand( double angle, double distance );
int clamp( int min, int val, int max );

int verbose_flag=0;
int tty_usb=0;

int main(int argc, char **argv) {
  int opt;
#ifdef HAVE_GETOPT_LONG
  int option_index;
#endif
  unsigned char buffer[256+3]="";
  char *tty=NULL;
  int irmode;
  int i;
  char name[32];
  int id;
  int err;
  pthread_t listenerThread;
  pthread_attr_t *tattr = NULL;
  int ports[32];
  int numPorts = 0;


  for( i = 0; i < sizeof(robotMaps)/sizeof(robotMaps[0]); i++ )
    {
      robotMaps[i].id = -1;
      robotMaps[i].name[0] = '\000';
    }
  
  while((opt=getopt_long(argc, argv, "t:i:p:r:v",
			 (struct option *)long_options, &option_index) )!=-1) {
    switch(opt) {
    case 'p':
      if ( numPorts < 32 )
	{
	  numPorts = addPort( ports, numPorts, optarg );
	}
      else
	{
	  fprintf( stderr, "Too many ports. I already registered %d ports\n", numPorts );
	}
      break;
    case 't':
      sscanf(optarg,"%s",buffer);
      break;
    case 'i':
      sscanf(optarg,"%x",&irmode);
      break;
    case 'r':
      err = extractRobotMapping( optarg, name, &id );
      if ( ! err )
	{
	  err = insertRobotMapping( robotMaps, sizeof( robotMaps )/ sizeof( robotMaps[0] ), name, id );
	  if ( err )
	    {
	      fprintf( stderr, "Failed to insert robot %s id %d mapping\n", name, id );
	    }
	}
      else
	{
	  fprintf( stderr, "Failed to extract robot mapping from %s\n", optarg );
	}
      break;
    case 'v':
      verbose_flag=1;
      break;
    }
  }           
  
  // load executable
  //      
  if ( argc-optind != 0 ) 
    {
      char *usage_string =
	"Options:\n"
	"  -t<comport>  , --tty=<comport>       set IR Tower com port <comport>\n"
#if defined(_WIN32)
	"  -t<usb>      , --tty=<usb>           set IR Tower USB mode \n"
#else
        "                                       (if \"usb\" is in the port, use USB mode)\n"
#endif
	"  -p port      , --port port           port number to listen on\n"
	"  -r robot=id  , --robot robot=id      robot to id mapping\n"
	"  -v           , --verbose             verbose mode\n"
	"Default COM port or USB support can be set using environment variable RCXTTY.\n"
	"Eg:\tset RCXTTY=COM2\n"
	"\tset RCXTTY=USB\n"
	"\n"
	;
      
      fprintf(stderr, "usage: %s [options]\n", argv[0]);
      fprintf(stderr, usage_string);
      
      return -1;
    }

  if ( robotMaps[0].id == -1 ) 
    {
      char const * defNames[8] = { "blue1", "yellow1", 
				   "blue2", "yellow2",
				   "blue3", "yellow3", 
				   "blue4", "yellow4" };
      int i;

      for( i = 0; i < 8; i++ )
	{
	  err = insertRobotMapping( robotMaps, sizeof( robotMaps )/ sizeof( robotMaps[0] ), 
				    ( char * ) defNames[i], i );
	  if ( err )
	    {
	      fprintf( stderr, "Failed to insert robot %s id %d mapping\n", defNames[i], i );
	    }      
	}
    }

  if ( numPorts == 0 )
    {
      ports[numPorts++] = DEFAULT_PORT;
    }

  // Moved tty device setting for a new option on command line
  if (buffer[0]) tty=buffer;
  if (!tty) tty = getenv(TTY_VARIABLE);
  if (!tty) tty = DEFAULTTTY;

  // Check if USB IR tower is selected.
#if defined(_WIN32)
  if (stricmp(tty, "usb")==0) {
	tty_usb = 1;
	if(verbose_flag)
	  {
	    fputs("\n\n Hary Mahesan - LEGO USB IR Tower Mode\n\n",stderr);
	  }
	tty="\\\\.\\legotower1"; // Set the correct usb tower if you have more than one (unlikely).
  }
#elif defined(LINUX) || defined(linux)
   /* If the tty string contains "usb", e.g. /dev/usb/lego0, we */
   /* assume it is the USB tower.  /dev/usb/lego0 is the default name of */
   /* the device in the LegoUSB (http://legousb.sourceforge.net) project. */
   /* If yours doesn't contain the "usb" string, just link it. */
   if (strstr(tty,"usb") !=0) {
       tty_usb=1;
       if (verbose_flag)
          fputs("\nC.P. Chan & Tyler Akins - USB IR Tower Mode for Linux.\n",stderr);
   }
#endif

  if (verbose_flag)
    fputs("\nCommand server initialized...\n",stderr);

  for( i = 0; i < numPorts; i++ )
    {
      if (pthread_create
	  (&listenerThread, 
	   tattr, 
	   listener,
	   &ports[i]) != 0)
	{
	  perror ("main pthread_create failed\n");
	}
    }

  LNPinit(tty); 

  for (;; )
    {
      sendMessage( currentCommand, 8 );
      usleep( 50000 ); 
    }
  return 0;
}

int addPort( int ports[], int numPorts, char * newPort )
{
  int p;
  int i;
  int found = 0;

  p = atoi( newPort );
  
  for( i = 0; i < numPorts; i++ )
    {
      if ( ports[i] == p )
	{
	  fprintf( stderr, "addPort: port p %d already defined as input\n", p );
	  found = 1;
	  break;
	}
    }
  if ( ! found ) 
    {
      ports[numPorts++] = p;
    }
  return numPorts;
}

int sendMessage( unsigned char volatile * buffer, int len )
{
  unsigned char message[10];
  int i;
  int found;
  static int idleCount = 0;

  if ( len == 8 )
    {
      memcpy( &message[1], buffer, len );
      memset( buffer, IGNORE_COMMAND, len );
      message[0] = START_BYTE;
      message[9] = STOP_MASK_BYTE | calcCheckSum( message, 9 );

      found = 0;
      for( i = 1; i < 9; i++)
	{
	  if ( message[i] != IGNORE_COMMAND )
	    {
	      found = 1;
	      break;
	    }
	}
      if ( ( found ) || ( idleCount < 20 ) )
	{
	  if ( found )
	    {
	      idleCount = 0;
	    }
	  else
	    {
	      idleCount++;
	    }

	  fprintf(stderr,"sendMessage " );
	  for( i = 0; i < 10; i++ )
	    {
	      fprintf( stderr,"%02x", message[i] );
	    }
	  fprintf( stderr,"\n" );
	  
	  if ( lnp_logical_write( message, 10 ) )
	    {
	      fprintf( stderr, "error downloading message\n" );
	      return -1;
	    }
	}
    }
  else
    {
      fprintf( stderr, "sendMessage message length must be %d\n", len );
    }
  return 0;
}



/*! parseMessage
 *! returns 0 on success
 */

int parseMessage( char * buffer, int len )
{
  int token;
  unsigned char commandCode;
  enum 
    {
      NAME = 1,
      NUM1,
      NUM2,
      S_NL,
      S_NL_NOERR,
    }
  state;
  double angle;
  double distance;
  int map;

  state = NAME;

  messageBuffer = buffer;
  messageLength = len;

  for( ;; )
    {
      token = yylex();
      fprintf(stderr, "parseRobotCommand Read in token %d %s\n", token, yytext );

      if ( token == 0 )
	{
	  if ( ( state != NAME )  && ( state != S_NL_NOERR ) )
	    {
	      fprintf( stderr, "listener: unexpected end of message\n" );
	    }
	  break;
	}
      
      switch( state )
	{
	case NAME:
	  if ( token == NAME )
	    {
	      if ( ( map = findRobotMapping( yytext ) ) != -1 )
		{
		  state = NUM1;
		}
	      else
		{
		  fprintf(stderr,"listener thread: no mapping for robot %s\n",
			  yytext);
		  state = S_NL_NOERR;
		}
	    }
	  break;
	case NUM1:
	  if ( token == HEXBYTE )
	    {
	      commandCode = ( unsigned char ) atoi( yytext );
	      if ( ( commandCode == START_BYTE ) ||
		   ( commandCode == STOP_MASK_BYTE ) )
		{
		  commandCode = IGNORE_COMMAND;
		}
	      state = S_NL;
	    }
	  else if ( token == NUMBER )
	    {
	      commandCode = atoi( yytext );
	      angle = atof( yytext );
	      state = NUM2;
	    }
	  else if ( token == KICKL )
	    {
	      commandCode = KICK_LEFT;
	      state = S_NL;
	    }
	  else if ( token == KICKR )
	    {
	      commandCode = KICK_RIGHT;
	      state = S_NL;
	    }
	  else if ( token == KICKS )
	    {
	      commandCode = KICK_STRAIGHT;
	      state = S_NL;
	    }
	  else
	    {
	      state = S_NL_NOERR;
	    }
	  break;
	case NUM2:
	  if ( token == NUMBER )
	    {
	      distance = atof( yytext );
	      commandCode = encodeCommand( angle, distance );
	      state = S_NL;
	    }	  
	  else if ( token == NEWLINE )
	    {
	    }
	  break;
	case S_NL:
	  if ( token == NEWLINE )
	    {
#ifdef DEBUG
	      fprintf(stderr, "setting currentCommand[%d] to %02x\n", map, commandCode );
#endif	      
	      currentCommand[ map ] = commandCode;
	      state = NAME;
	    }
	  else
	    {
	      fprintf( stderr, "parseRobotCommand missing newline\n" );
	      state = S_NL_NOERR;
	    }
  	  break;
	case S_NL_NOERR:
	  if ( token == NEWLINE )
	    {
	      state = NAME;
	    }
  	  break;
	}      
    }
  return 0;
}

/*! findRobotMapping
 *!
 */

int findRobotMapping( char const * name )
{
  int i;
  int map = -1;

  for( i = 0; i < sizeof( robotMaps ) / sizeof( robotMaps[0] ) ; i++ )
    {
      if ( ! strcmp( name, robotMaps[i].name ) )
	{
	  map = robotMaps[i].id;
	  break;
	}
    }  
#ifdef DEBUG
  fprintf(stderr, "findRobotMapping: returns %d for robot %s\n", map, name );
#endif
  return map;
}


/*! listener thread function
 *!
 */

void listener( int * port )
{
  char buf[256];
  int count;
  int udpSocket;

  memset (buf, IGNORE_COMMAND, sizeof (buf));

  udpSocket = initUDPBroadcastSocket( * port );

  for( ;; )
    {
      count = recvfrom (udpSocket, buf, sizeof (buf), 0, 0, 0);
      if ( count == -1 )
	{
	  break;
	}

      parseMessage( buf, count );
      
    }
}


/*! extract robot mapping from a string.
 *!
 */

int extractRobotMapping( char const * arg, char * name, int * id )
{
  char * out;
  char const * p, * cp;

  cp = strchr( arg, '=' );
  if ( cp != NULL )
    {
      if ( sscanf( cp + 1 , "%i", id ) != 1 )
	{
	  fprintf( stderr, "extractRobotMapping: unable to extract id %s\n", cp + 1 );
	  return 1;
	}
      p = arg;
      out = name;
      while ( ( p != cp ) && ( *p != '\0' ) )
	{
	  *out++ = *p++;
	}
      *out++ = '\0';
    }
  else
    {
      fprintf( stderr, "extractRobotMapping: unable to extract robot mapping from %s\n", arg );
      return 1;
    }
  return 0;
}

/*! insertRobotMapping: insert a mapping for a robot into the map\n
 *!
 */

int insertRobotMapping( struct RobotMapping * maps, int numMaps, char * name, int id )
{
  int i;
  int inserted = 0;

  for( i = 0 ; i < numMaps; i++ )
    {
      if ( maps[i].id == id )
	{
	  fprintf( stderr, "insertRobotMapping: duplicate assignment for robots (old=%s) and (new=%s) with id %d\n",
		   ( char *)&(maps[i].name), name, id );
	  return 1;
	}
      if ( maps[i].id == -1 )
	{
	  strcpy( maps[i].name, name );
	  maps[i].id = id;
	  inserted = 1;
	  break;
	}
    }
  if ( ! inserted )
    {
      fprintf( stderr, "insertRobotMapping: insertion of robot %s with %d id failed too many robots\n", name, id );
      return 1;
    }
  return 0;
}

/*! blocking I/R write.
 *! return number of bytes written, or negative for error.
 */
int lnp_logical_write(const void *data, size_t length) 
{
// With Win32 we are using Blocking Write by default
#if !defined(_WIN32)
  fd_set fds;

  // wait for transmission
  //
  do {
    FD_ZERO(&fds);
    FD_SET(rcxFD(),&fds);
  } while(select(rcxFD()+1,NULL,&fds,NULL,NULL)<1);
#endif

  // transmit
  //
  return (mywrite(rcxFD(), data, length)!=length);
}

void LNPinit(const char *tty) {
  struct timeval timeout,now;
  long diff;
  int i;
  int err;

#if !defined(_WIN32)
  unsigned char buffer[256];
#endif
    
  // initialize RCX communications
  //
  if (verbose_flag) fputs("opening tty...\n", stderr);
#ifdef CONF_LNP_FAST
  rcxInit(tty, 1);
#else
  rcxInit(tty, 0);
#endif

  if (rcxFD() == BADFILE) {
    myperror("opening tty");
    exit(1);
  }

#if defined(LINUX) || defined(linux)
  if (tty_usb == 0) {
#endif
     
  // wait for IR to settle
  // 
  gettimeofday(&timeout,0);
  do {    
    usleep(100000);
    gettimeofday(&now,0);
    diff=1000000*(now.tv_sec  - timeout.tv_sec ) +
         	  now.tv_usec - timeout.tv_usec;
      
  } while(diff < 100000);
#if defined(LINUX) || defined(linux)
  }
#endif

#if defined(LINUX) || defined(linux)
  legousbapi_debug=1;
  for(i = 3; i >= 1; i-- )
    {
      if ( ( err = legousbapi_set_parm( rcxFD(), 2, i ) ) == 0x04 )
	{
	  fprintf( stderr, "Selected transmission range %d\n", i );
	  break;
	}
      else
	{
	  fprintf( stderr, "Selecting transmission range %d failed (err=%d)\n", i, err );
	}
    }
  legousbapi_debug=0;			   
#endif

#if defined(_WIN32)
  PurgeComm(rcxFD(), PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
#else
#if defined(LINUX) || defined(linux)
   if (tty_usb == 0)
#endif
  read(rcxFD(),buffer,256);
#endif
}

int initUDPBroadcastSocket( int port )
{
  int s;
  struct sockaddr_in src;
  int temp;

  if (-1 == (s = socket (AF_INET, SOCK_DGRAM, 0)))
    {
      perror ("socket open failed");
    }

  src.sin_family = AF_INET;
  src.sin_port = htons (port);
  src.sin_addr.s_addr = htonl (INADDR_ANY);
  //  src.sin_addr.s_addr = htonl ( 192 << 24 | 168 << 16 | 25 << 8 | 102 );

  temp = 1;
  if (setsockopt (s, SOL_SOCKET, SO_REUSEADDR, &temp, sizeof ( temp ) )
      == -1)
    {
      perror ("setsockopt error");
    }

  if (-1 == bind (s, (struct sockaddr *) &src, sizeof (src)))
    {
      perror ("bind error");
    }
  return s;
}

unsigned int calcCheckSum( unsigned char * message, int length )
{
  unsigned char a = 0xff;
  unsigned char t;
  
  do {
    t = *message++;
    a += t;
    length--;
  } while (length > 0);

  return a & 0x07;
}

unsigned char encodeCommand( double angle, double distance )
{
  double const angleScale = 1.0/M_PI*180.0;         /* radians to 2 deg. */
  double const distanceScale = 1.0/10.0;            /* mm to cm */
  int const maxAngle = 14;
  int const maxDistance = 7;
  int t;
  int d;
  
  t = (int) rint( angle * angleScale );
  d = (int) rint( distance * distanceScale );
  
  t = clamp(-maxAngle, t, maxAngle );
  d = clamp(-maxDistance, d, maxDistance );

#ifdef DEBUG
  fprintf( stderr, "encodeCommand(angle=%f, distance=%f)=%d %d\n", angle, distance, t, d );
#endif
  return ( ( t/2 << 4 ) & 0xf0 ) | ( d & 0x0f );
}

int clamp( int min, int val, int max )
{
  if ( val < min )
    {
      val = min;
    }
  else if ( val > max ) 
    {
      val = max;
    }
  return val;
}
